module nl.netsoj.chris.blog.interfaces.http;

import std.experimental.logger;

import vibe.d;

import nl.netsoj.chris.blog.interfaces.indieauth;
import nl.netsoj.chris.blog.model.article;
import nl.netsoj.chris.blog.model.page;
import nl.netsoj.chris.blog.model.project;
import nl.netsoj.chris.blog.cache;
import nl.netsoj.chris.blog.constants;

/**
 * Output types for the content.
 */
enum OutputType {
	HTML,
	MARKDOWN
}

immutable string MIME_MARKDOWN = "text/markdown";

immutable Duration CACHE_TIME = days(16);

/**
 * Get's the document type for the given slug based on extension
 * and returns the slug without extension and the document type. Also removes the extension from the
 * slug.
 */
private OutputType getOutputType(ref string slug) {
	if (slug.endsWith(".md")) {
		slug = chomp(slug, ".md");
		return OutputType.MARKDOWN;
	} else if (slug.endsWith(".html")){
		// If explicitly asking for HTML, we'll return HTML
		slug = chomp(slug, ".html");
		return OutputType.HTML;
	} else {
		// If in the future, for any reason, we no longer use HTML
		// this allows to us to keep the current urls with an option
		// to change the output in the future.
		return OutputType.HTML;
	}
}

void addCachingHeader(bool publicCache = true)(ref HTTPServerResponse res) {
	string header = "";
	static if (publicCache) {
		header ~= "public";
	} else {
		header ~= "private";
	}
	header ~= ", max-age=" ~ to!string(CACHE_TIME.total!"seconds");
	res.headers["Cache-Control"] = header;
}



struct TranslateContext {
	import std.typetuple;
	
	alias languages = TypeTuple!("en_GB", "nl_NL");
	mixin translationModule!"mijnblog";
	static string determineLanguage(scope HTTPServerRequest req) {
		if ("lang" !in req.query) return req.determineLanguageByHeader(languages); // default behaviour using "Accept-Language" header
		return req.query.get("lang", "en_GB");
	}
}

/**
 * Generates boilerplate code for a single response.
 * params:
 *     arrayName = The name of the associative array to take the items from.
 *     templateName = The name of the template to render.
 */
string singleResponseMixin(string arrayName, string templateName) {
	return `string slug = req.params["slug"];
	OutputType outputType = getOutputType(slug);

	enforceHTTP(slug in ` ~ arrayName ~ `, HTTPStatus.notFound, "Page not found");
	auto content = ` ~ arrayName ~ `[slug];
	switch(outputType) with (OutputType) {
	case MARKDOWN:
		res.writeBody(content.contentSource, MIME_MARKDOWN);
		break;
	default:
	case HTML:
		render!("` ~ templateName ~ `", content);
		break;
	}`;
}

@translationContext!TranslateContext
class MijnBlog {

public:
	
	this() {
		m_indieAuth = new IndieAuth();
	}
	
	/** 
	 * IndieAuth subinterface
	 */
	@safe
	@path("/indieweb")
	@property IndieAuth indieweb() { return m_indieAuth; };
	
	/**
	 * Generates response for /posts/:slug and /palen/:slug.
	 */
	@path("/posts/:slug")
	void getArticleSingle(string _slug, HTTPServerRequest req, HTTPServerResponse res) {
		//getSingle!(Article, "pages/article.dt")(articles, req, res);
		mixin(singleResponseMixin("articles", "pages/article.dt"));
	}

	/**
	* Generates response for /posts and /palen
	*/
	@path("/posts/")
	void getArticleOverview(HTTPServerRequest req, HTTPServerResponse res) {
		addCachingHeader(res);
		Article[] articleList = articles.sortedList;
		render!("pages/article-list.dt", articleList);
	}

	/**
	* Generates response for /projects and /projecten
	*/
	@path("/projects/")
	void getProjectOverview(HTTPServerRequest req, HTTPServerResponse res) {
		addCachingHeader(res);
		Project[] projectList = projects.sortedList;
		render!("pages/project-list.dt", projectList);
	}
	
	/**
	* Generate response for a project page
	*/
	@path("/projects/:slug")
	void getProject(HTTPServerRequest req, HTTPServerResponse res) {
		res.headers["Cache-Control"] = "public";
		mixin(singleResponseMixin("projects", "pages/project.dt"));
	}
	/**
	* Generate response for a page
	*/
	@path("/:slug")
	void getPage(HTTPServerRequest req, HTTPServerResponse res) {
		addCachingHeader(res);
		mixin(singleResponseMixin("pages", "pages/page.dt"));
	}

	@path("/")
	void getIndexPage(HTTPServerRequest req, HTTPServerResponse res) {
		addCachingHeader(res);
		// If no slug is supplied, it will be adjusted to "index"
		req.params.addField("slug", "index");
		res.headers.addField("Link", "<" ~ Constants.SITE_URL ~ "/indieweb/auth>; rel=\"authorization_endpoint\"");
		res.headers.addField("Link", "<" ~ Constants.SITE_URL ~ "/indieweb/token>; rel=\"token_endpoint\"");
		res.headers.addField("Link", "<" ~ Constants.SITE_URL ~ "/indieweb/micropub>; rel=\"micropub\"");
		mixin(singleResponseMixin("pages", "pages/page.dt"));
	}
private:
	IndieAuth m_indieAuth;
}

/**
 * Generates response whenever an error occurs.
 */
@safe
void errorPage(HTTPServerRequest req, HTTPServerResponse res, HTTPServerErrorInfo error) {
	//render!("pages/error.dt", error)(res);
	import std.conv;
	res.writeBody(text("Error ", error.code, ": ", error.message, "\n\n", error.debugMessage), "text/plain");
}

@trusted
void startHTTPServer() {
	HTTPServerSettings settings = new HTTPServerSettings;
	settings.bindAddresses = ["0.0.0.0"];
	settings.port = 3465;
	settings.serverString = "zeg ik lekker niet";
	settings.errorPageHandler = toDelegate(&errorPage);
	settings.keepAliveTimeout = dur!"seconds"(60);
	debug {
		settings.accessLogToConsole = true;
	}
	HTTPFileServerSettings fSettings = new HTTPFileServerSettings;
	fSettings.maxAge = days(16);
	
	URLRouter router = new URLRouter;
	router.get("/static/*", serveStaticFiles("./public/", fSettings));
	router.registerWebInterface(new MijnBlog);
	
	listenHTTP(settings, router);
	foreach(route; router.getAllRoutes()) {
		infof("Path: %s", route);
	}
}
